/*
 *  linux/fs/minix/file.c
 *
 *  Copyright (C) 1991, 1992 Linus Torvalds
 *
 *  minix regular file handling primitives
 */

#include <linux/sched.h>
#include <linux/minix_fs.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

/* DebEvan */
#include <asm/current.h>
#include <asm/uaccess.h>
#include <asm/system.h>

#define    NBUF    32

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

#include <linux/fs.h>

/* DebEvan */
#include "active.h"
#include <linux/string.h>

/* DebEvan */
#define MAXLINENO 10000
static unsigned int LINENO = 1;
static int sbDoingLogging = 0 ;

static ssize_t minix_file_write(struct file *, const char *, size_t, loff_t *);
static ssize_t minix_file_open(struct inode *, struct file *);
static ssize_t minix_file_release(struct inode *, struct file *);

/*
 * We have mostly NULLs here: the current defaults are OK for
 * the minix filesystem.
 */
static struct file_operations minix_file_operations = {
    NULL,            /* lseek - default */
    generic_file_read,    /* read */
    minix_file_write,    /* write */
    NULL,            /* readdir - bad */
    NULL,            /* poll - default */
    NULL,            /* ioctl - default */
    generic_file_mmap,    /* mmap */
    minix_file_open,    /* DebEvan: open */
    NULL,            /* flush */
    minix_file_release,    /* DebEvan: release */
    minix_sync_file        /* fsync */
};

struct inode_operations minix_file_inode_operations = {
    &minix_file_operations,    /* default file operations */
    NULL,            /* create */
    NULL,            /* lookup */
    NULL,            /* link */
    NULL,            /* unlink */
    NULL,            /* symlink */
    NULL,            /* mkdir */
    NULL,            /* rmdir */
    NULL,            /* mknod */
    NULL,            /* rename */
    NULL,            /* readlink */
    NULL,            /* follow_link */
    generic_readpage,    /* readpage */
    NULL,            /* writepage */
    minix_bmap,        /* bmap */
    minix_truncate,        /* truncate */
    NULL            /* permission */
};

/* 22-Apr DebEvan */
int minix_write_log(int fileOp, char* psfileName)
{
    char    buf[MAX_AUDIT_BUFSIZE+1];
    char    sFileOp[12];
    char    cAck;
    long    time;
    int     uid, len, len1, err=0;
    mm_segment_t old_fs;

    /* Is logging, wherever it goes, wanted? */
    if (!DO_LOGGING) {
        return err ;
    }

    /* If we got this far, we are doing logging, either in the
     * kernel or at the user level. If we are doing user level
     * logging, the kernel may or may not waiting for acknowledgement
     * from the listener active daemon.
     */

    /* To avoid loops, only log activity if we are not already
     * logging activity.
      */
    if (sbDoingLogging == 0) {
        sbDoingLogging = 1;

        /* begin to build the output string */
        switch (fileOp) {
            case ACTIVE_FILEOP_CREATE:
                sprintf(sFileOp, "CREATE");
                break;
            case ACTIVE_FILEOP_OPEN:
                sprintf(sFileOp, "OPEN");
                break;
            case ACTIVE_FILEOP_CLOSE:
                sprintf(sFileOp, "CLOSE");
                break;
            case ACTIVE_FILEOP_UNLINK:
                sprintf(sFileOp, "UNLINK");
                break;
            default:
                sprintf(sFileOp, "Unknown op");
        }

        /* get the user id */
        /* borrowed from sys_getuid() in kernel/sched.c */
        uid = current->uid ;

        /* get the current time */
        time = CURRENT_TIME ;

        /* build the output string */
        /**** need to guard against writing too many characters!!! ****/
        memset((void*)buf, '\0', (MAX_AUDIT_BUFSIZE+1));

        /* DebEvan:    added LINENO to the active.log record
         * Still need to decide if and when the line number
         * should be reset.  Currently, it is reset to one
         * when the active module is remade.
         * It would probably be better to reset when the
         * active.log is removed.  But, maybe not either.
         */
        /* build the audit string */
        sprintf(buf, "%d;%s;%s;%d;%ld\n",
                (LINENO++ % MAXLINENO), sFileOp, psfileName, uid, time);
        len = strlen(buf);

        /* Write the audit file to either directly to the log file
         * or to the named pipe.
         */
        if (!gfd_kout_err) {
            /* write to the file */
            old_fs = get_fs();
            set_fs(KERNEL_DS);

            len1 = fd_kout->f_op->write(fd_kout, buf, len, &fd_kout->f_pos);
            if (len1 != len) {
                err = -1 ;
            }
            /* is the kernel supposed to wait for an acknowledgement? */
            if (DO_ACKNOWLEDGEMENT) {
                len1 = fd_l2k->f_op->read(fd_l2k, &cAck, 1, &fd_l2k->f_pos);
//printk("Kernel heard back: %c.\n", cAck);
            }

            set_fs(old_fs);
        }

        /* clear the flag to allow other file ops to be logged */
        sbDoingLogging = 0 ;
    }
    return err;
}

/* DebEvan */
static ssize_t minix_file_release(struct inode * inode, struct file * filp)
{
    int err ;
    char namebuf[NAMEBUFLEN];

    err = minix_write_log(ACTIVE_FILEOP_CLOSE, 
        d_path(filp->f_dentry, namebuf, NAMEBUFLEN) );
    return 0;
}

/* DebEvan */
static ssize_t minix_file_open(struct inode * inode, struct file * filp)
{
    int err ;
    char namebuf[NAMEBUFLEN];

    err = minix_write_log(ACTIVE_FILEOP_OPEN, 
        d_path(filp->f_dentry, namebuf, NAMEBUFLEN) );
    return 0;
}

static ssize_t minix_file_write(struct file * filp, const char * buf,
                size_t count, loff_t *ppos)
{
    struct inode * inode = filp->f_dentry->d_inode;
    off_t pos;
    ssize_t written, c;
    struct buffer_head * bh;
    char * p;

    if (!inode) {
        printk("minix_file_write: inode = NULL\n");
        return -EINVAL;
    }
    if (!S_ISREG(inode->i_mode)) {
        printk("minix_file_write: mode = %07o\n",inode->i_mode);
        return -EINVAL;
    }
    if (filp->f_flags & O_APPEND)
        pos = inode->i_size;
    else
        pos = *ppos;
    written = 0;
    while (written < count) {
        bh = minix_getblk(inode,pos/BLOCK_SIZE,1);
        if (!bh) {
            if (!written)
                written = -ENOSPC;
            break;
        }
        c = BLOCK_SIZE - (pos % BLOCK_SIZE);
        if (c > count-written)
            c = count-written;
        if (c != BLOCK_SIZE && !buffer_uptodate(bh)) {
            ll_rw_block(READ, 1, &bh);
            wait_on_buffer(bh);
            if (!buffer_uptodate(bh)) {
                brelse(bh);
                if (!written)
                    written = -EIO;
                break;
            }
        }
        p = (pos % BLOCK_SIZE) + bh->b_data;
        c -= copy_from_user(p,buf,c);
        if (!c) {
            brelse(bh);
            if (!written)
                written = -EFAULT;
            break;
        }
        update_vm_cache(inode, pos, p, c);
        mark_buffer_uptodate(bh, 1);
        mark_buffer_dirty(bh, 0);
        brelse(bh);
        pos += c;
        written += c;
        buf += c;
    }
    if (pos > inode->i_size)
        inode->i_size = pos;
    inode->i_mtime = inode->i_ctime = CURRENT_TIME;
    *ppos = pos;
    mark_inode_dirty(inode);
    return written;
}
